; dat     segment
        org  100h
;         assume  cs:dat,ds:dat,es:dat
; code    PROC    ; 16.16 fixpoint Mandelbrot figure drawing MCGA 3/1995
; .386

	; 9/2003: translated to NASM dialect, did some palette finetuning,
	; added USE192 define and some defines and macros to improve the
	; general readability of the code.

	; (c) By Eric Auer 1995-2003. Feel free to use this code in any other
	; open source project. You only have to leave this boring disclaimer
	; in here. This draws a nice Mandelbrot figure for ME, but I do not
	; say that it will do anything useful (and nothing harmful) for YOU!

%define USE192 1	; only use 189 colors, not 251

; --------------------

%imacro MIDDLE32 3	; 3 parameters
	mov [tmp+0],%1	; put low  32bits
	mov [tmp+4],%2	; put high 32bits
	mov %3,[tmp+2]	; fetch middle 32bits
%endmacro		; similar to "shrd" ...,16 bits

%imacro NORMALIZE 1	; take middle 32bit of EDX:EAX after IMUL 32x32->64bit
	; MIDDLE32 eax,edx, %1
	shrd eax,edx,16	; 386 double shift operation: edx:eax shr 16 -> eax
	mov %1,eax
%endmacro

%imacro SETZERO 1
	xor %1,%1
%endmacro

; --------------------

        mov ax,0013h	; set 320x200 8bit/pixel graphics mode
        int 10h
        call makepal	; set up color palette

        cli		; disable interrupts
        and esp,0fff8h
        pushad		; save all registers
        push es		; save es
        mov [resp],esp	; even ESP will be used as a variable, so save it!


        mov ax,0a000h
        mov es,ax	; load graphics buffer segment into ES

%define xx ebx	; ebx/ebp=xx/yy (square of...)
%define yy ebp
%define x  esi	; esi/edi=x/y
%define y  edi
%define x0 ecx	; ecx/esp=x0/y0
%define y0 esp
		; eax/edx=Temp.

        SETZERO edx		; clear high 32bit of 64/32 -> 32bit divide
        mov eax,[rxmax]
        sub eax,[rxmin]
        idiv dword [rxpix]
        mov [RXINC],eax		; Xincrement is (Xmax-Xmin)/Xresolution

        SETZERO edx		; clear high 32bit of 64/32 -> 32bit divide
        mov eax,[rymax]
        sub eax,[rymin]
        idiv dword [rypix]
        mov [RYINC],eax		; Yincrement is (Ymax-Ymin)/Yresolution

        mov x0,[rxmin]		; delta value, depends on coordinates
        mov y0,[rymin]		; (Mandelbrot formula: p = p*p + delta)
	; (Julia formula: delta fixed, but not initial p, which was 0 for M.)

pixel:				; prepare for next pixel
        SETZERO xx		; squares from last iteration, initially 0
        SETZERO yy
        SETZERO x		; value initially (0,0)
        SETZERO y

        SETZERO ax
        mov [rcount],ax		; counter initially 0

; INNER LOOP BEGINS HERE

iterate:
        mov eax,x
        imul y
	NORMALIZE y
        add y,y
        add y,y0		; y = y0 + (2*x*y)  (im=im0+2*re*im)

        mov x,xx
        sub x,yy
        add x,x0		; x = x0 + xx-yy (re=re0+oldREsqr-oldIMsqr)

        mov eax,x
        imul x
	NORMALIZE xx		; xx = square(x)

        mov eax,y
        imul y
	NORMALIZE yy		; yy = square(y)

        mov ax,[rcount]
        inc ax
        mov [rcount],ax
        cmp ax,[rcmax]		; reached maximum count?
        jnb  inside             ; JNB (then leave the loop early)

        mov eax,xx		; if maximum count not yet reached...
        add eax,yy
        cmp eax,000040000h      ; (xx + yy) < 4.0 ?
        jle  iterate            ; JLE (loop while below 4.0)

; INNER LOOP ENDS HERE

inside:				; we reached maximum count or
				; xx + yy became more than 4.0
        mov ax,[rcount]		; count of iterations, used as COLOR

        mov di,[rpixel]		; PIXEL address in RAM
        stosb			; mov [es:di],ax ; inc edi	; SET PIXEL
        mov [rpixel],di		; PIXEL address in RAM

        add x0,[RXINC]		; set x0 for next pixel
        cmp x0,[rxmax]		; reached right edge?
        jl  near pixel		; if not, loop

        mov x0,[rxmin]		; set x0 to left edge again
        add y0,[RYINC]		; set y0 for next pixel row
        cmp y0,[rymax]		; reached bottom?
        jl  near pixel		; if not, loop

; --------------------

        mov esp,[resp]	; restore esp
        pop es		; restore es
        popad		; restore registers
        sti		; allow interrupts

        mov ax,0000h	; wait for a key
        int 16h
        mov ax,0003h	; go to 80x25 text mode
        int 10h
        mov ax,4c00h	; exit with errorlevel 0
        int 21h

; --------------------

	align 4

tmp	dd 0,0		; 64bit buffer for "take middle 32 bit"

resp    dd 0fff8h	; saved esp

; --------------------

rxmin   dd 0fffe0000h   ;-2.0   left edge X
rxmax   dd 000008000h   ; 0.5  right edge X

rxpix   dd 320		; screen X resolution
RXINC   dd 999		; fixpoint position increment per pixel

rymin   dd 0fffecccdh   ;-1.2    top edge Y
rymax   dd 000013333h   ; 1.2 bottom edge Y

rypix   dd 200		; screen Y resolution
RYINC   dd 999		; fixpoint position increment per pixel

rpixel  dw 0            ; 0 (screen buffer address of pixel)

rcount  dw 0		; count of iterations
%ifdef USE192
rcmax	dw 189		; 3*63
%else
rcmax   dw 252          ; 4*63 (maximum count of iterations, at most 255)
%endif

; --------------------

makepal:		; set up some color ramps for a nice display
        call warte
        SETZERO bx	; start with black color 0
        SETZERO cx	; start with black
        mov dx,3c8h	; DAC color number
        mov al,bl
        out dx,al	; start with 0
        inc dx		; DAC color data

rv:     call setinc	; first part: black ... red (0..63)
        inc bh		; +R
        cmp bl,63
        jnz rv          ; (+R)

grv:    call setinc	; second part: red ... yellow/(green) (63..126)
        inc cl		; +G
        		; dec bh	; -R (for ... green)
	cmp bl,126
        jnz grv         ; (for ... green: +G -R / for ... yellow: +G)

%ifdef USE192
bgv:	call setinc	; third part: yellow/(green) ... white (126..189)
        inc ch		; +B
			; inc bh	; +R (from green)
	cmp bl,189
        jnz bgv         ; (from green: +R +B / from yellow: +B)
%else
bgv:    call setinc		; third part: yellow/(green) ... blue (..189)
        inc ch			; +B
        dec cl			; -G
	dec bh			; -R (from yellow)
	cmp bl,189
        jnz bgv         	; (from green: +B -G / from yellow: +B -G -R)

bv:     call setinc		; fourth part: blue ... black/white (..252)
        			; dec ch	; -B (for ... black)
	inc cl			; +G (for ... white)
	inc bh			; +R (for ... white)
        cmp bl,252
        jnz bv          	; (-B or +R +G)
%endif

        mov bh,3fh	; override last color by a static value
        mov cx,3f3fh
        call setinc     ; last color is 2/3 grey (light blue: bh=36 cx=3f36)
        ret

; --------------------

setinc: inc bl		; output BH, CL and CH after another to DX. inc BL.
        mov al,bh	; (used to send a color definition to the DAC).
        out dx,al	; BH is Red, CL is Green, CH is Blue.
        jmp short nop1
nop1:   mov al,cl
        out dx,al
        jmp short nop2
nop2:   mov al,ch
        out dx,al
        ret

; --------------------

warte:  SETZERO cx	; wait for a nice retrace moment
        mov ah,10
        mov dx,3dah
wa_lp:  in al,dx	; wait until 3da test 8 enters 0 state
        test al,8
        loopnz wa_lp
        jz fnd0
        dec ah		; allow 10 x 64k wait cycles!
        jnz wa_lp
fnd0:   in al,dx	; wait until 3da test 8 leaves 0 state
        test al,8
        loopz fnd0
fnd1:   ret

; --------------------

; code    ENDP
; dat     ENDS
; END     CODE
